# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from builtins import object
from future import standard_library
standard_library.install_aliases()

import unittest

from duplicity import dup_time
from duplicity import file_naming
from duplicity import log
from duplicity import globals
from duplicity import util
from . import UnitTestCase


class Test36(UnitTestCase):
    def test_base36(self):
        u"""Test conversion to/from base 36"""
        numlist = [0, 1, 10, 1313, 34233, 872338, 2342889,
                   134242234, 1204684368, 34972382455]
        for n in numlist:
            b = file_naming.to_base36(n)
            assert file_naming.from_base36(b) == n, (b, n)


class FileNamingBase(object):
    u"""Holds file naming test functions, for use in subclasses"""
    def test_basic(self):
        u"""Check get/parse cycle"""
        dup_time.setprevtime(10)
        dup_time.setcurtime(20)

        file_naming.prepare_regex(force=True)
        filename = file_naming.get(u"inc", volume_number=23)
        log.Info(u"Inc filename: " + util.fsdecode(filename))
        pr = file_naming.parse(filename)
        assert pr and pr.type == u"inc", pr
        assert pr.start_time == 10
        assert pr.end_time == 20
        assert pr.volume_number == 23
        assert not pr.partial

        filename = file_naming.get(u"full-sig")
        log.Info(u"Full sig filename: " + util.fsdecode(filename))
        pr = file_naming.parse(filename)
        assert pr.type == u"full-sig"
        assert pr.time == 20
        assert not pr.partial

        filename = file_naming.get(u"new-sig")
        pr = file_naming.parse(filename)
        assert pr.type == u"new-sig"
        assert pr.start_time == 10
        assert pr.end_time == 20
        assert not pr.partial

    def test_suffix(self):
        u"""Test suffix (encrypt/compressed) encoding and generation"""
        file_naming.prepare_regex(force=True)
        filename = file_naming.get(u"inc", manifest=1, gzipped=1)
        pr = file_naming.parse(filename)
        assert pr and pr.compressed == 1
        assert pr.manifest

        filename2 = file_naming.get(u"full", volume_number=23, encrypted=1)
        pr = file_naming.parse(filename2)
        assert pr and pr.encrypted == 1
        assert pr.volume_number == 23

    def test_more(self):
        u"""More file_parsing tests"""
        file_naming.prepare_regex(force=True)
        pr = file_naming.parse(globals.file_prefix + globals.file_prefix_signature + b"dns.h112bi.h14rg0.st.g")
        assert pr, pr
        assert pr.type == u"new-sig"
        assert pr.end_time == 1029826800

        if not globals.short_filenames:
            pr = file_naming.parse(globals.file_prefix + globals.file_prefix_signature + b"duplicity-new-signatures.2002-08-18T00:04:30-07:00.to.2002-08-20T00:00:00-07:00.sigtar.gpg")
            assert pr, pr
            assert pr.type == u"new-sig"
            assert pr.end_time == 1029826800

        pr = file_naming.parse(globals.file_prefix + globals.file_prefix_signature + b"dfs.h5dixs.st.g")
        assert pr, pr
        assert pr.type == u"full-sig"
        assert pr.time == 1036954144, repr(pr.time)

    def test_partial(self):
        u"""Test addition of partial flag"""
        file_naming.prepare_regex(force=True)
        pr = file_naming.parse(globals.file_prefix + globals.file_prefix_signature + b"dns.h112bi.h14rg0.st.p.g")
        assert pr, pr
        assert pr.partial
        assert pr.type == u"new-sig"
        assert pr.end_time == 1029826800

        if not globals.short_filenames:
            pr = file_naming.parse(globals.file_prefix + globals.file_prefix_signature + b"duplicity-new-signatures.2002-08-18T00:04:30-07:00.to.2002-08-20T00:00:00-07:00.sigtar.part.gpg")
            assert pr, pr
            assert pr.partial
            assert pr.type == u"new-sig"
            assert pr.end_time == 1029826800

        pr = file_naming.parse(globals.file_prefix + globals.file_prefix_signature + b"dfs.h5dixs.st.p.g")
        assert pr, pr
        assert pr.partial
        assert pr.type == u"full-sig"
        assert pr.time == 1036954144, repr(pr.time)


class FileNamingLong(UnitTestCase, FileNamingBase):
    u"""Test long filename parsing and generation"""
    def setUp(self):
        super(FileNamingLong, self).setUp()
        self.set_global(u'short_filenames', 0)


class FileNamingShort(UnitTestCase, FileNamingBase):
    u"""Test short filename parsing and generation"""
    def setUp(self):
        super(FileNamingShort, self).setUp()
        self.set_global(u'short_filenames', 1)


class FileNamingPrefixes(UnitTestCase, FileNamingBase):
    u"""Test filename parsing and generation with prefixes"""
    def setUp(self):
        super(FileNamingPrefixes, self).setUp()
        self.set_global(u'file_prefix', b"global-")
        self.set_global(u'file_prefix_manifest', b"mani-")
        self.set_global(u'file_prefix_signature', b"sign-")
        self.set_global(u'file_prefix_archive', b"arch-")


if __name__ == u"__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from builtins import map
from builtins import object
from builtins import range
from future import standard_library
standard_library.install_aliases()

import io
import unittest

from duplicity import diffdir
from duplicity import patchdir
from duplicity import log  # @UnusedImport
from duplicity import selection
from duplicity import tarfile  # @UnusedImport
from duplicity import librsync  # @UnusedImport
from duplicity.path import *  # @UnusedWildImport
from . import UnitTestCase


class PatchingTest(UnitTestCase):
    u"""Test patching"""
    def setUp(self):
        super(PatchingTest, self).setUp()
        self.unpack_testfiles()

    def copyfileobj(self, infp, outfp):
        u"""Copy in fileobj to out, closing afterwards"""
        blocksize = 32 * 1024
        while 1:
            buf = infp.read(blocksize)
            if not buf:
                break
            outfp.write(buf)
        assert not infp.close()
        assert not outfp.close()

    def test_total(self):
        u"""Test cycle on dirx"""
        self.total_sequence([u'testfiles/dir1',
                             u'testfiles/dir2',
                             u'testfiles/dir3'])

    def get_sel(self, path):
        u"""Get selection iter over the given directory"""
        return selection.Select(path).set_iter()

    def total_sequence(self, filelist):
        u"""Test signatures, diffing, and patching on directory list"""
        assert len(filelist) >= 2
        sig = Path(u"testfiles/output/sig.tar")
        diff = Path(u"testfiles/output/diff.tar")
        seq_path = Path(u"testfiles/output/sequence")
        new_path, old_path = None, None  # set below in for loop

        # Write initial full backup to diff.tar
        for dirname in filelist:
            old_path, new_path = new_path, Path(dirname)
            if old_path:
                sigblock = diffdir.DirSig(self.get_sel(seq_path))
                diffdir.write_block_iter(sigblock, sig)
                deltablock = diffdir.DirDelta(self.get_sel(new_path),
                                              sig.open(u"rb"))
            else:
                deltablock = diffdir.DirFull(self.get_sel(new_path))
            diffdir.write_block_iter(deltablock, diff)

            patchdir.Patch(seq_path, diff.open(u"rb"))
            # print "#########", seq_path, new_path
            assert seq_path.compare_recursive(new_path, 1)

    def test_block_tar(self):
        u"""Test building block tar from a number of files"""
        def get_fileobjs():
            u"""Return iterator yielding open fileobjs of tar files"""
            for i in range(1, 4):
                p = Path(u"testfiles/blocktartest/test%d.tar" % i)
                fp = p.open(u"rb")
                yield fp
                fp.close()

        tf = patchdir.TarFile_FromFileobjs(get_fileobjs())
        namelist = []
        for tarinfo in tf:
            namelist.append(tarinfo.name)
        for i in range(1, 6):
            assert (u"tmp/%d" % i) in namelist, namelist

    def test_doubledot_hole(self):
        u"""Test for the .. bug that lets tar overwrite parent dir"""

        def make_bad_tar(filename):
            u"""Write attack tarfile to filename"""
            tf = tarfile.TarFile(name=filename, mode=u"w")

            # file object will be empty, and tarinfo will have path
            # "snapshot/../warning-security-error"
            assert not os.system(u"cat /dev/null >testfiles/output/file")
            path = Path(b"testfiles/output/file")
            path.index = (b"diff", b"..", b"warning-security-error")
            ti = path.get_tarinfo()
            fp = io.StringIO(u"")
            tf.addfile(ti, fp)

            tf.close()

        make_bad_tar(b"testfiles/output/bad.tar")
        os.mkdir(u"testfiles/output/temp")

        self.assertRaises(patchdir.PatchDirException, patchdir.Patch,
                          Path(u"testfiles/output/temp"),
                          open(u"testfiles/output/bad.tar", u"rb"))
        assert not Path(u"testfiles/output/warning-security-error").exists()


class index(object):
    u"""Used below to test the iter collation"""
    def __init__(self, index):
        self.index = index


class CollateItersTest(UnitTestCase):
    def setUp(self):
        super(CollateItersTest, self).setUp()
        self.unpack_testfiles()

    def test_collate(self):
        u"""Test collate_iters function"""
        indicies = [index(i) for i in [0, 1, 2, 3]]
        helper = lambda i: indicies[i]

        makeiter1 = lambda: iter(indicies)
        makeiter2 = lambda: map(helper, [0, 1, 3])
        makeiter3 = lambda: map(helper, [1, 2])

        outiter = patchdir.collate_iters([makeiter1(), makeiter2()])
        assert Iter.equal(outiter,
                          iter([(indicies[0], indicies[0]),
                                (indicies[1], indicies[1]),
                                (indicies[2], None),
                                (indicies[3], indicies[3])]))

        assert Iter.equal(patchdir.collate_iters([makeiter1(),
                                                 makeiter2(),
                                                 makeiter3()]),
                          iter([(indicies[0], indicies[0], None),
                                (indicies[1], indicies[1], indicies[1]),
                                (indicies[2], None, indicies[2]),
                                (indicies[3], indicies[3], None)]), 1)

        assert Iter.equal(patchdir.collate_iters([makeiter1(), iter([])]),
                          map(lambda i: (i, None), indicies))
        assert Iter.equal(map(lambda i: (i, None), indicies),
                          patchdir.collate_iters([makeiter1(), iter([])]))

    def test_tuple(self):
        u"""Test indexed tuple"""
        i = patchdir.IndexedTuple((1, 2, 3), (u"a", u"b"))
        i2 = patchdir.IndexedTuple((), (u"hello", u"there", u"how are you"))

        assert i[0] == u"a"
        assert i[1] == u"b"
        assert i2[1] == u"there"
        assert len(i) == 2 and len(i2) == 3
        assert i2 < i, i2 < i

    def test_tuple_assignment(self):
        a, b, c = patchdir.IndexedTuple((), (1, 2, 3))
        assert a == 1
        assert b == 2
        assert c == 3


class TestInnerFuncs(UnitTestCase):
    u"""Test some other functions involved in patching"""
    def setUp(self):
        super(TestInnerFuncs, self).setUp()
        self.unpack_testfiles()
        self.check_output()

    def check_output(self):
        u"""Make sure testfiles/output exists"""
        out = Path(u"testfiles/output")
        if not (out.exists() and out.isdir()):
            out.mkdir()
        self.out = out

    def snapshot(self):
        u"""Make a snapshot ROPath, permissions 0o600"""
        ss = self.out.append(u"snapshot")
        fout = ss.open(u"wb")
        fout.write(b"hello, world!")
        assert not fout.close()
        ss.chmod(0o600)
        ss.difftype = u"snapshot"
        return ss

    def get_delta(self, old_buf, new_buf):
        u"""Return delta buffer from old to new"""
        sigfile = librsync.SigFile(io.BytesIO(old_buf))
        sig = sigfile.read()
        assert not sigfile.close()

        deltafile = librsync.DeltaFile(sig, io.BytesIO(new_buf))
        deltabuf = deltafile.read()
        assert not deltafile.close()
        return deltabuf

    def delta1(self):
        u"""Make a delta ROPath, permissions 0o640"""
        delta1 = self.out.append(u"delta1")
        fout = delta1.open(u"wb")
        fout.write(self.get_delta(b"hello, world!",
                                  b"aonseuth aosetnuhaonsuhtansoetuhaoe"))
        assert not fout.close()
        delta1.chmod(0o640)
        delta1.difftype = u"diff"
        return delta1

    def delta2(self):
        u"""Make another delta ROPath, permissions 0o644"""
        delta2 = self.out.append(u"delta1")
        fout = delta2.open(u"wb")
        fout.write(self.get_delta(b"aonseuth aosetnuhaonsuhtansoetuhaoe",
                                  b"3499 34957839485792357 458348573"))
        assert not fout.close()
        delta2.chmod(0o644)
        delta2.difftype = u"diff"
        return delta2

    def deleted(self):
        u"""Make a deleted ROPath"""
        deleted = self.out.append(u"deleted")
        assert not deleted.exists()
        deleted.difftype = u"deleted"
        return deleted

    def test_normalize(self):
        u"""Test normalizing a sequence of diffs"""
        ss = self.snapshot()
        d1 = self.delta1()
        d2 = self.delta2()
        de = self.deleted()

        seq1 = [ss, d1, d2]
        seq2 = [ss, d1, d2, de]
        seq3 = [de, ss, d1, d2]
        seq4 = [de, ss, d1, d2, ss]
        seq5 = [de, ss, d1, d2, ss, d1, d2]

        def try_seq(input_seq, correct_output_seq):
            normed = patchdir.normalize_ps(input_seq)
            assert normed == correct_output_seq, (normed, correct_output_seq)

        try_seq(seq1, seq1)
        try_seq(seq2, [de])
        try_seq(seq3, seq1)
        try_seq(seq4, [ss])
        try_seq(seq5, seq1)

    def test_patch_seq2ropath(self):
        u"""Test patching sequence"""
        def testseq(seq, perms, buf):
            result = patchdir.patch_seq2ropath(seq)
            assert result.getperms() == perms, (result.getperms(), perms)
            fout = result.open(u"rb")
            contents = fout.read()
            assert not fout.close()
            assert contents == buf, (contents, buf)

        ids = u"%d:%d" % (os.getuid(), os.getgid())

        testseq([self.snapshot()], (u"%s 600" % ids), b"hello, world!")
        testseq([self.snapshot(), self.delta1()], (u"%s 640" % ids),
                b"aonseuth aosetnuhaonsuhtansoetuhaoe")
        testseq([self.snapshot(), self.delta1(), self.delta2()], (u"%s 644" % ids),
                b"3499 34957839485792357 458348573")


if __name__ == u"__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

u"""Generate and process backup statistics"""
from __future__ import division

from builtins import zip
from builtins import map
from builtins import str
from builtins import object

import re
import time
import os

from duplicity import dup_time


class StatsException(Exception):
    pass


class StatsObj(object):
    u"""Contains various statistics, provide string conversion functions"""
    # used when quoting files in get_stats_line
    space_regex = re.compile(u" ")

    stat_file_attrs = (u'SourceFiles',
                       u'SourceFileSize',
                       u'NewFiles',
                       u'NewFileSize',
                       u'DeletedFiles',
                       u'ChangedFiles',
                       u'ChangedFileSize',
                       u'ChangedDeltaSize',
                       u'DeltaEntries',
                       u'RawDeltaSize')
    stat_misc_attrs = (u'Errors',
                       u'TotalDestinationSizeChange')
    stat_time_attrs = (u'StartTime',
                       u'EndTime',
                       u'ElapsedTime')
    stat_attrs = ((u'Filename',) + stat_time_attrs +
                  stat_misc_attrs + stat_file_attrs)

    # Below, the second value in each pair is true iff the value
    # indicates a number of bytes
    stat_file_pairs = ((u'SourceFiles', False),
                       (u'SourceFileSize', True),
                       (u'NewFiles', False),
                       (u'NewFileSize', True),
                       (u'DeletedFiles', False),
                       (u'ChangedFiles', False),
                       (u'ChangedFileSize', True),
                       (u'ChangedDeltaSize', True),
                       (u'DeltaEntries', False),
                       (u'RawDeltaSize', True))

    # This is used in get_byte_summary_string below
    byte_abbrev_list = ((1024 * 1024 * 1024 * 1024, u"TB"),
                        (1024 * 1024 * 1024, u"GB"),
                        (1024 * 1024, u"MB"),
                        (1024, u"KB"))

    def __init__(self):
        u"""Set attributes to None"""
        for attr in self.stat_attrs:
            self.__dict__[attr] = None

    def get_stat(self, attribute):
        u"""Get a statistic"""
        return self.__dict__[attribute]

    def set_stat(self, attr, value):
        u"""Set attribute to given value"""
        self.__dict__[attr] = value

    def increment_stat(self, attr):
        u"""Add 1 to value of attribute"""
        self.__dict__[attr] += 1

    def get_total_dest_size_change(self):
        u"""Return total destination size change

        This represents the total increase in the size of the
        duplicity destination directory, or None if not available.

        """
        return 0  # this needs to be re-done for duplicity

    def get_stats_line(self, index, use_repr=1):
        u"""Return one line abbreviated version of full stats string"""
        file_attrs = [str(self.get_stat(a)) for a in self.stat_file_attrs]
        if not index:
            filename = u"."
        else:
            filename = os.path.join(*index)
            if use_repr:
                # use repr to quote newlines in relative filename, then
                # take of leading and trailing quote and quote spaces.
                filename = self.space_regex.sub(u"\\\\x20", repr(filename))
                n = 1
                if filename[0] == u'u':
                    n = 2
                filename = filename[n:-1]
        return u" ".join([filename, ] + file_attrs)

    def set_stats_from_line(self, line):
        u"""Set statistics from given line"""
        def error():
            raise StatsException(u"Bad line '%s'" % line)
        if line[-1] == u"\n":
            line = line[:-1]
        lineparts = line.split(u" ")
        if len(lineparts) < len(self.stat_file_attrs):
            error()
        for attr, val_string in zip(self.stat_file_attrs,
                                    lineparts[-len(self.stat_file_attrs):]):
            try:
                val = int(val_string)
            except ValueError:
                try:
                    val = float(val_string)
                except ValueError:
                    error()
            self.set_stat(attr, val)
        return self

    def get_stats_string(self):
        u"""Return extended string printing out statistics"""
        return u"%s%s%s" % (self.get_timestats_string(),
                            self.get_filestats_string(),
                            self.get_miscstats_string())

    def get_timestats_string(self):
        u"""Return portion of statistics string dealing with time"""
        timelist = []
        if self.StartTime is not None:
            timelist.append(u"StartTime %.2f (%s)\n" %  # pylint: disable=bad-string-format-type
                            (self.StartTime, dup_time.timetopretty(self.StartTime)))
        if self.EndTime is not None:
            timelist.append(u"EndTime %.2f (%s)\n" %  # pylint: disable=bad-string-format-type
                            (self.EndTime, dup_time.timetopretty(self.EndTime)))
        if self.ElapsedTime or (self.StartTime is not None and
                                self.EndTime is not None):
            if self.ElapsedTime is None:
                self.ElapsedTime = self.EndTime - self.StartTime
            timelist.append(u"ElapsedTime %.2f (%s)\n" %
                            (self.ElapsedTime, dup_time.inttopretty(self.ElapsedTime)))
        return u"".join(timelist)

    def get_filestats_string(self):
        u"""Return portion of statistics string about files and bytes"""
        def fileline(stat_file_pair):
            u"""Return zero or one line of the string"""
            attr, in_bytes = stat_file_pair
            val = self.get_stat(attr)
            if val is None:
                return u""
            if in_bytes:
                return u"%s %s (%s)\n" % (attr, val,
                                          self.get_byte_summary_string(val))
            else:
                return u"%s %s\n" % (attr, val)

        return u"".join(map(fileline, self.stat_file_pairs))

    def get_miscstats_string(self):
        u"""Return portion of extended stat string about misc attributes"""
        misc_string = u""
        tdsc = self.TotalDestinationSizeChange
        if tdsc is not None:
            misc_string += (u"TotalDestinationSizeChange %s (%s)\n" %
                            (tdsc, self.get_byte_summary_string(tdsc)))
        if self.Errors is not None:
            misc_string += u"Errors %d\n" % self.Errors
        return misc_string

    def get_byte_summary_string(self, byte_count):
        u"""Turn byte count into human readable string like "7.23GB" """
        if byte_count < 0:
            sign = u"-"
            byte_count = -byte_count
        else:
            sign = u""

        for abbrev_bytes, abbrev_string in self.byte_abbrev_list:
            if byte_count >= abbrev_bytes:
                # Now get 3 significant figures
                abbrev_count = float(byte_count) / abbrev_bytes
                if abbrev_count >= 100:
                    precision = 0
                elif abbrev_count >= 10:
                    precision = 1
                else:
                    precision = 2
                return u"%s%%.%df %s" % (sign, precision, abbrev_string) \
                       % (abbrev_count,)
        byte_count = round(byte_count)
        if byte_count == 1:
            return sign + u"1 byte"
        else:
            return u"%s%d bytes" % (sign, byte_count)

    def get_stats_logstring(self, title):
        u"""Like get_stats_string, but add header and footer"""
        header = u"--------------[ %s ]--------------" % title
        footer = u"-" * len(header)
        return u"%s\n%s%s\n" % (header, self.get_stats_string(), footer)

    def set_stats_from_string(self, s):
        u"""Initialize attributes from string, return self for convenience"""
        def error(line):
            raise StatsException(u"Bad line '%s'" % line)

        for line in s.split(u"\n"):
            if not line:
                continue
            line_parts = line.split()
            if len(line_parts) < 2:
                error(line)
            attr, value_string = line_parts[:2]
            if attr not in self.stat_attrs:
                error(line)
            try:
                try:
                    val1 = int(value_string)
                except ValueError:
                    val1 = None
                val2 = float(value_string)
                if val1 == val2:
                    self.set_stat(attr, val1)  # use integer val
                else:
                    self.set_stat(attr, val2)  # use float
            except ValueError:
                error(line)
        return self

    def write_stats_to_path(self, path):
        u"""Write statistics string to given path"""
        fin = path.open(u"w")
        fin.write(self.get_stats_string())
        assert not fin.close()

    def read_stats_from_path(self, path):
        u"""Set statistics from path, return self for convenience"""
        fp = path.open(u"r")
        self.set_stats_from_string(fp.read())
        assert not fp.close()
        return self

    def stats_equal(self, s):
        u"""Return true if s has same statistics as self"""
        assert isinstance(s, StatsObj)
        for attr in self.stat_file_attrs:
            if self.get_stat(attr) != s.get_stat(attr):
                return None
        return 1

    def set_to_average(self, statobj_list):
        u"""Set self's attributes to average of those in statobj_list"""
        for attr in self.stat_attrs:
            self.set_stat(attr, 0)
        for statobj in statobj_list:
            for attr in self.stat_attrs:
                if statobj.get_stat(attr) is None:
                    self.set_stat(attr, None)
                elif self.get_stat(attr) is not None:
                    self.set_stat(attr, statobj.get_stat(attr) +
                                  self.get_stat(attr))

        # Don't compute average starting/stopping time
        self.StartTime = None
        self.EndTime = None

        for attr in self.stat_attrs:
            if self.get_stat(attr) is not None:
                self.set_stat(attr,
                              self.get_stat(attr) / float(len(statobj_list)))
        return self

    def get_statsobj_copy(self):
        u"""Return new StatsObj object with same stats as self"""
        s = StatsObj()
        for attr in self.stat_attrs:
            s.set_stat(attr, self.get_stat(attr))
        return s


class StatsDeltaProcess(StatsObj):
    u"""Keep track of statistics during DirDelta process"""
    def __init__(self):
        u"""StatsDeltaProcess initializer - zero file attributes"""
        StatsObj.__init__(self)
        for attr in StatsObj.stat_file_attrs:
            self.__dict__[attr] = 0
        self.Errors = 0
        self.StartTime = time.time()
        self.files_changed = []

    def add_new_file(self, path):
        u"""Add stats of new file path to statistics"""
        filesize = path.getsize()
        self.SourceFiles += 1
        # SourceFileSize is added-to incrementally as read
        self.NewFiles += 1
        self.NewFileSize += filesize
        self.DeltaEntries += 1
        self.add_delta_entries_file(path, b'new')

    def add_changed_file(self, path):
        u"""Add stats of file that has changed since last backup"""
        filesize = path.getsize()
        self.SourceFiles += 1
        # SourceFileSize is added-to incrementally as read
        self.ChangedFiles += 1
        self.ChangedFileSize += filesize
        self.DeltaEntries += 1
        self.add_delta_entries_file(path, b'changed')

    def add_deleted_file(self, path):
        u"""Add stats of file no longer in source directory"""
        self.DeletedFiles += 1  # can't add size since not available
        self.DeltaEntries += 1
        self.add_delta_entries_file(path, b'deleted')

    def add_unchanged_file(self, path):
        u"""Add stats of file that hasn't changed since last backup"""
        filesize = path.getsize()
        self.SourceFiles += 1
        self.SourceFileSize += filesize

    def close(self):
        u"""End collection of data, set EndTime"""
        self.EndTime = time.time()

    def add_delta_entries_file(self, path, action_type):
        if path.isreg():
            self.files_changed.append((path.get_relative_path(), action_type))

    def get_delta_entries_file(self):
        return self.files_changed



from __future__ import print_function
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from builtins import object
from future import standard_library
standard_library.install_aliases()

import unittest

from duplicity.path import *  # @UnusedWildImport
from duplicity import diffdir
from duplicity import selection
from duplicity import util
from duplicity import tarfile  # @Reimport
from . import UnitTestCase


class DDTest(UnitTestCase):
    u"""Test functions in diffdir.py"""
    def setUp(self):
        super(DDTest, self).setUp()
        self.unpack_testfiles()

    def copyfileobj(self, infp, outfp):
        u"""Copy in fileobj to out, closing afterwards"""
        blocksize = 32 * 1024
        while 1:
            buf = infp.read(blocksize)
            if not buf:
                break
            outfp.write(buf)
        assert not infp.close()
        assert not outfp.close()

    def testsig(self):
        u"""Test producing tar signature of various file types"""
        select = selection.Select(Path(u"testfiles/various_file_types"))
        select.set_iter()
        sigtar = diffdir.SigTarBlockIter(select)
        diffdir.write_block_iter(sigtar, u"testfiles/output/sigtar")

        i = 0
        for tarinfo in tarfile.TarFile(u"testfiles/output/sigtar", u"r"):  # @UnusedVariable
            i += 1
        assert i >= 5, u"There should be at least 5 files in sigtar"

    def empty_diff_schema(self, dirname):
        u"""Given directory name, make sure can tell when nothing changes"""
        select = selection.Select(Path(dirname))
        select.set_iter()
        sigtar = diffdir.SigTarBlockIter(select)
        diffdir.write_block_iter(sigtar, u"testfiles/output/sigtar")

        sigtar_fp = open(u"testfiles/output/sigtar", u"rb")
        select2 = selection.Select(Path(dirname))
        select2.set_iter()
        diffdir.write_block_iter(diffdir.DirDelta(select2, sigtar_fp),
                                 u"testfiles/output/difftar")

        size = os.stat(u"testfiles/output/difftar").st_size
        assert size == 0 or size == 10240, size  # 10240 is size of one record
        if size != 0:
            fin = open(u"testfiles/output/difftar", u"rb")
            diff_buf = fin.read()
            assert not fin.close()
            assert diff_buf == b'\0' * 10240

    def test_empty_diff(self):
        u"""Test producing a diff against same sig; should be len 0"""
        self.empty_diff_schema(u"testfiles/various_file_types")

        select = selection.Select(Path(u"testfiles/various_file_types"))
        select.set_iter()
        sigtar = diffdir.SigTarBlockIter(select)
        diffdir.write_block_iter(sigtar, u"testfiles/output/sigtar")

        sigtar_fp = open(u"testfiles/output/sigtar", u"rb")
        select2 = selection.Select(Path(u"testfiles/various_file_types"))
        select2.set_iter()
        diffdir.write_block_iter(diffdir.DirDelta(select2, sigtar_fp),
                                 u"testfiles/output/difftar")

        size = os.stat(u"testfiles/output/difftar").st_size  # @UnusedVariable

    def test_empty_diff2(self):
        u"""Test producing diff against directories of special files"""
        self.empty_diff_schema(u"testfiles/special_cases/neg_mtime")
        self.empty_diff_schema(u"testfiles/special_cases/no_uname")

    def test_diff(self):
        u"""Test making a diff"""
        sel1 = selection.Select(Path(u"testfiles/dir1"))
        diffdir.write_block_iter(diffdir.SigTarBlockIter(sel1.set_iter()),
                                 u"testfiles/output/dir1.sigtar")

        sigtar_fp = open(u"testfiles/output/dir1.sigtar", u"rb")
        sel2 = selection.Select(Path(u"testfiles/dir2"))
        delta_tar = diffdir.DirDelta(sel2.set_iter(), sigtar_fp)
        diffdir.write_block_iter(delta_tar,
                                 u"testfiles/output/dir1dir2.difftar")

        changed_files = [u"diff/changeable_permission",
                         u"diff/regular_file",
                         u"snapshot/symbolic_link/",
                         u"deleted/deleted_file",
                         u"snapshot/directory_to_file",
                         u"snapshot/file_to_directory/"]
        for tarinfo in tarfile.TarFile(u"testfiles/output/dir1dir2.difftar",
                                       u"r"):
            tiname = util.get_tarinfo_name(tarinfo)
            if tiname in changed_files:
                changed_files.remove(tiname)
        assert not changed_files, (u"Following files not found:\n"
                                   + u"\n".join(changed_files))

    def test_diff2(self):
        u"""Another diff test - this one involves multivol support
        (requires rdiff to be installed to pass)"""
        sel1 = selection.Select(Path(u"testfiles/dir2"))
        diffdir.write_block_iter(diffdir.SigTarBlockIter(sel1.set_iter()),
                                 u"testfiles/output/dir2.sigtar")

        sigtar_fp = open(u"testfiles/output/dir2.sigtar", u"rb")
        sel2 = selection.Select(Path(u"testfiles/dir3"))
        delta_tar = diffdir.DirDelta(sel2.set_iter(), sigtar_fp)
        diffdir.write_block_iter(delta_tar,
                                 u"testfiles/output/dir2dir3.difftar")

        buffer = b""
        tf = tarfile.TarFile(u"testfiles/output/dir2dir3.difftar", u"r")
        for tarinfo in tf:
            if tarinfo.name.startswith(r"multivol_diff/"):
                buffer += tf.extractfile(tarinfo).read()
        assert 3000000 < len(buffer) < 4000000
        fout = open(u"testfiles/output/largefile.delta", u"wb")
        fout.write(buffer)
        fout.close()
        assert not os.system(u"rdiff patch testfiles/dir2/largefile "
                             u"testfiles/output/largefile.delta "
                             u"testfiles/output/largefile.patched")
        dir3large = open(u"testfiles/dir3/largefile", u"rb").read()
        patchedlarge = open(u"testfiles/output/largefile.patched", u"rb").read()
        assert dir3large == patchedlarge

    def test_dirdelta_write_sig(self):
        u"""Test simultaneous delta and sig generation

        Generate signatures and deltas of dirs1, 2, 3, 4 and compare
        those produced by DirDelta_WriteSig and other methods.

        """
        deltadir1 = Path(u"testfiles/output/dir.deltatar1")  # @UnusedVariable
        deltadir2 = Path(u"testfiles/output/dir.deltatar2")  # @UnusedVariable
        cur_full_sigs = Path(u"testfiles/output/fullsig.dir1")

        cur_dir = Path(u"testfiles/dir1")
        get_sel = lambda cur_dir: selection.Select(cur_dir).set_iter()
        diffdir.write_block_iter(diffdir.SigTarBlockIter(get_sel(cur_dir)),
                                 cur_full_sigs)

        sigstack = [cur_full_sigs]
        for dirname in [u'dir2', u'dir3', u'dir4']:
            # print "Processing ", dirname
            old_dir = cur_dir  # @UnusedVariable
            cur_dir = Path(u"testfiles/" + dirname)

            old_full_sigs = cur_full_sigs
            cur_full_sigs = Path(u"testfiles/output/fullsig." + dirname)

            delta1 = Path(u"testfiles/output/delta1." + dirname)
            delta2 = Path(u"testfiles/output/delta2." + dirname)
            incsig = Path(u"testfiles/output/incsig." + dirname)

            # Write old-style delta to deltadir1
            diffdir.write_block_iter(diffdir.DirDelta(get_sel(cur_dir),
                                                      old_full_sigs.open(u"rb")),
                                     delta1)

            # Write new signature and delta to deltadir2 and sigdir2, compare
            block_iter = diffdir.DirDelta_WriteSig(
                get_sel(cur_dir),
                [p.open(u"rb") for p in sigstack],
                incsig.open(u"wb"))
            sigstack.append(incsig)
            diffdir.write_block_iter(block_iter, delta2)

            # print delta1.name, delta2.name
            compare_tar(delta1.open(u"rb"), delta2.open(u"rb"))
            assert not os.system(u"cmp %s %s" % (delta1.uc_name, delta2.uc_name))

            # Write old-style signature to cur_full_sigs
            diffdir.write_block_iter(diffdir.SigTarBlockIter(get_sel(cur_dir)),
                                     cur_full_sigs)

    def test_combine_path_iters(self):
        u"""Test diffdir.combine_path_iters"""
        class Dummy(object):
            def __init__(self, index, other=None):
                self.index = index
                self.other = other

            def __repr__(self):
                return u"(%s %s)" % (self.index, self.other)

        def get_iter1():
            yield Dummy(())
            yield Dummy((1,))
            yield Dummy((1, 5), 2)

        def get_iter2():
            yield Dummy((), 2)
            yield Dummy((1, 5))
            yield Dummy((2,))

        def get_iter3():
            yield Dummy((), 3)
            yield Dummy((2,), 1)

        result = diffdir.combine_path_iters([get_iter1(),
                                             get_iter2(),
                                             get_iter3()])
        elem1 = next(result)
        assert elem1.index == () and elem1.other == 3, elem1
        elem2 = next(result)
        assert elem2.index == (1,) and elem2.other is None, elem2
        elem3 = next(result)
        assert elem3.index == (1, 5) and elem3.other is None
        elem4 = next(result)
        assert elem4.index == (2,) and elem4.other == 1
        try:
            elem5 = next(result)
        except StopIteration:
            pass
        else:
            assert 0, elem5


def compare_tar(tarfile1, tarfile2):
    u"""Compare two tarfiles"""
    tf1 = tarfile.TarFile(u"none", u"r", tarfile1)
    tf2 = tarfile.TarFile(u"none", u"r", tarfile2)
    tf2_iter = iter(tf2)

    for ti1 in tf1:
        try:
            ti2 = next(tf2_iter)
        except StopIteration:
            assert 0, (u"Premature end to second tarfile, "
                       u"ti1.name = %s" % ti1.name)
        # print "Comparing ", ti1.name, ti2.name
        assert tarinfo_eq(ti1, ti2), u"%s %s" % (ti1.name, ti2.name)
        if ti1.size != 0:
            fp1 = tf1.extractfile(ti1)
            buf1 = fp1.read()
            fp1.close()
            fp2 = tf2.extractfile(ti2)
            buf2 = fp2.read()
            fp2.close()
            assert buf1 == buf2
    try:
        ti2 = next(tf2_iter)
    except StopIteration:
        pass
    else:
        assert 0, (u"Premature end to first tarfile, "
                   u"ti2.name = %s" % ti2.name)

    tarfile1.close()
    tarfile2.close()


def tarinfo_eq(ti1, ti2):
    if ti1.name != ti2.name:
        print(u"Name:", ti1.name, ti2.name)
        return 0
    if ti1.size != ti2.size:
        print(u"Size:", ti1.size, ti2.size)
        return 0
    if ti1.mtime != ti2.mtime:
        print(u"Mtime:", ti1.mtime, ti2.mtime)
        return 0
    if ti1.mode != ti2.mode:
        print(u"Mode:", ti1.mode, ti2.mode)
        return 0
    if ti1.type != ti2.type:
        print(u"Type:", ti1.type, ti2.type)
        return 0
    if ti1.issym() or ti1.islnk():
        if ti1.linkname != ti2.linkname:
            print(u"Linkname:", ti1.linkname, ti2.linkname)
            return 0
    if ti1.uid != ti2.uid or ti1.gid != ti2.gid:
        print(u"IDs:", ti1.uid, ti2.uid, ti1.gid, ti2.gid)
        return 0
    if ti1.uname != ti2.uname or ti1.gname != ti2.gname:
        print(u"Owner names:", ti1.uname, ti2.uname, ti1.gname, ti2.gname)
        return 0
    return 1

if __name__ == u"__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

u"""
Error/exception classes that do not fit naturally anywhere else.
"""

from duplicity import log


class DuplicityError(Exception):
    pass


class UserError(DuplicityError):
    u"""
    Subclasses use this in their inheritance hierarchy to signal that
    the error is a user generated one, and that it is therefore
    typically unsuitable to display a full stack trace.
    """
    pass


class NotSupported(DuplicityError):
    u"""
    Exception raised when an action cannot be completed because some
    particular feature is not supported by the environment.
    """
    pass


class ConflictingScheme(DuplicityError):
    u"""
    Raised to indicate an attempt was made to register a backend for a
    scheme for which there is already a backend registered.
    """
    pass


class InvalidBackendURL(UserError):
    u"""
    Raised to indicate a URL was not a valid backend URL.
    """
    pass


class UnsupportedBackendScheme(InvalidBackendURL, UserError):
    u"""
    Raised to indicate that a backend URL was parsed successfully as a
    URL, but was not supported.
    """
    def __init__(self, url):
        InvalidBackendURL.__init__(self,
                                   (u"scheme not supported in url: %s" % (url,)))
        self.url = url


class BackendException(DuplicityError):
    u"""
    Raised to indicate a backend specific problem.
    """
    def __init__(self, msg, code=log.ErrorCode.backend_error):
        super(BackendException, self).__init__(msg)
        self.code = code


class FatalBackendException(BackendException):
    u"""
    Raised to indicate a backend failed fatally.
    """
    pass


class TemporaryLoadException(BackendException):
    u"""
    Raised to indicate a temporary issue on the backend.
    Duplicity should back off for a bit and try again.
    """
    pass



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2019 Francesco Magno
# Copyright 2019 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import range
from builtins import object

import os
import os.path

import duplicity.backend
from duplicity import path
from duplicity import log
from duplicity.errors import BackendException
from duplicity import util


class RcloneBackend(duplicity.backend.Backend):

    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)
        self.rclone_cmd = u"rclone"
        self.parsed_url = parsed_url
        self.remote_path = self.parsed_url.path

        try:
            rc, o, e = self.subprocess_popen(
                self.rclone_cmd + u" --version")
        except Exception:
            log.FatalError(u"rclone not found: please install rclone", log.ErrorCode.backend_error)

        if parsed_url.path.startswith(u"//"):
            self.remote_path = self.remote_path[2:].replace(u":/", u":", 1)

        self.remote_path = util.fsdecode(self.remote_path)

    def _get(self, remote_filename, local_path):
        remote_filename = util.fsdecode(remote_filename)
        local_pathname = util.fsdecode(local_path.name)
        temp_dir = os.path.dirname(local_pathname)
        commandline = u"%s copy %s/%s %s" % (
            self.rclone_cmd, self.remote_path, remote_filename, temp_dir)
        rc, o, e = self.subprocess_popen(commandline)
        if rc != 0:
            if os.path.isfile(os.path.join(temp_dir, remote_filename)):
                os.remove(os.path.join(temp_dir, remote_filename))
            raise BackendException(e.split(u'\n')[0])
        os.rename(os.path.join(temp_dir, remote_filename), local_pathname)

    def _put(self, source_path, remote_filename):
        source_pathname = util.fsdecode(source_path.name)
        remote_filename = util.fsdecode(remote_filename)
        temp_dir = util.fsdecode(os.path.dirname(source_pathname))
        temp_filename = os.path.basename(source_pathname)
        os.rename(source_pathname, os.path.join(temp_dir, remote_filename))
        commandline = u"%s copy --include %s %s %s" % (
            self.rclone_cmd, remote_filename, temp_dir, self.remote_path)
        rc, o, e = self.subprocess_popen(commandline)
        if rc != 0:
            os.rename(os.path.join(temp_dir, remote_filename), source_pathname)
            raise BackendException(e.split(u'\n')[0])
        os.rename(os.path.join(temp_dir, remote_filename), source_pathname)

    def _list(self):
        filelist = []
        commandline = u"%s ls %s" % (
            self.rclone_cmd, self.remote_path)
        rc, o, e = self.subprocess_popen(commandline)
        if rc != 0:
            if e.endswith(u"not found\n"):
                return filelist
            else:
                raise BackendException(e.split(u'\n')[0])
        if not o:
            return filelist
        return [util.fsencode(x.split()[-1]) for x in o.split(u'\n') if x]

    def _delete(self, remote_filename):
        remote_filename = util.fsdecode(remote_filename)
        commandline = u"%s delete --drive-use-trash=false --include %s %s" % (
            self.rclone_cmd, remote_filename, self.remote_path)
        rc, o, e = self.subprocess_popen(commandline)
        if rc != 0:
            raise BackendException(e.split(u'\n')[0])


duplicity.backend.register_backend(u"rclone", RcloneBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from builtins import range
from future import standard_library
standard_library.install_aliases()

import os
import unittest

from duplicity import path
from . import CmdError, FunctionalTestCase


class FinalTest(FunctionalTestCase):
    u"""
    Test backup/restore using duplicity binary
    """
    def runtest(self, dirlist, backup_options=[], restore_options=[]):
        u"""Run backup/restore test on directories in dirlist"""
        assert len(dirlist) >= 1

        # Back up directories to local backend
        current_time = 100000
        self.backup(u"full", dirlist[0], current_time=current_time,
                    options=backup_options)
        for new_dir in dirlist[1:]:
            current_time += 100000
            self.backup(u"inc", new_dir, current_time=current_time,
                        options=backup_options)

        # Restore each and compare them
        for i in range(len(dirlist)):
            dirname = dirlist[i]
            current_time = 100000 * (i + 1)
            self.restore(time=current_time, options=restore_options)
            self.check_same(dirname, u"testfiles/restore_out")
            self.verify(dirname,
                        time=current_time, options=restore_options)

    def check_same(self, filename1, filename2):
        u"""Verify two filenames are the same"""
        path1, path2 = path.Path(filename1), path.Path(filename2)
        assert path1.compare_recursive(path2, verbose=1)

    def test_basic_cycle(self, backup_options=[], restore_options=[]):
        u"""Run backup/restore test on basic directories"""
        self.runtest([u"testfiles/dir1",
                      u"testfiles/dir2",
                      u"testfiles/dir3"],
                     backup_options=backup_options,
                     restore_options=restore_options)

        # Test restoring various sub files
        for filename, time, dir in [(u'symbolic_link', 99999, u'dir1'),
                                    (u'directory_to_file', 100100, u'dir1'),
                                    (u'directory_to_file', 200100, u'dir2'),
                                    (u'largefile', 300000, u'dir3')]:
            self.restore(filename, time, options=restore_options)
            self.check_same(u'testfiles/%s/%s' % (dir, filename),
                            u'testfiles/restore_out')
            self.verify(u'testfiles/%s/%s' % (dir, filename),
                        file_to_verify=filename, time=time,
                        options=restore_options)

    def test_asym_cycle(self):
        u"""Like test_basic_cycle but use asymmetric encryption and signing"""
        backup_options = [u"--encrypt-key", self.encrypt_key1,
                          u"--sign-key", self.sign_key]
        restore_options = [u"--encrypt-key", self.encrypt_key1,
                           u"--sign-key", self.sign_key]
        self.test_basic_cycle(backup_options=backup_options,
                              restore_options=restore_options)

    def test_asym_with_hidden_recipient_cycle(self):
        u"""Like test_basic_cycle but use asymmetric encryption (hiding key id) and signing"""
        backup_options = [u"--hidden-encrypt-key", self.encrypt_key1,
                          u"--sign-key", self.sign_key]
        restore_options = [u"--hidden-encrypt-key", self.encrypt_key1,
                           u"--sign-key", self.sign_key]
        self.test_basic_cycle(backup_options=backup_options,
                              restore_options=restore_options)

    def test_single_regfile(self):
        u"""Test backing and restoring up a single regular file"""
        self.runtest([u"testfiles/various_file_types/regular_file"])

    def test_empty_backup(self):
        u"""Make sure backup works when no files change"""
        self.backup(u"full", u"testfiles/empty_dir")
        self.backup(u"inc", u"testfiles/empty_dir")

    def test_long_filenames(self):
        u"""Test backing up a directory with long filenames in it"""
        # Note that some versions of ecryptfs (at least through Ubuntu 11.10)
        # have a bug where they treat the max path segment length as 143
        # instead of 255.  So make sure that these segments don't break that.
        lf_dir = path.Path(u"testfiles/long_filenames")
        if lf_dir.exists():
            lf_dir.deltree()
        lf_dir.mkdir()
        lf1 = lf_dir.append(u"AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA")
        lf1.mkdir()
        lf2 = lf1.append(u"BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB")
        lf2.mkdir()
        lf3 = lf2.append(u"CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC")
        lf3.mkdir()
        lf4 = lf3.append(u"DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD")
        lf4.touch()
        lf4_1 = lf3.append(u"SYMLINK--------------------------------------------------------------------------------------------")
        os.symlink(u"SYMLINK-DESTINATION-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------", lf4_1.name)
        lf4_1.setdata()
        assert lf4_1.issym()
        lf4_2 = lf3.append(u"DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD")
        fp = lf4_2.open(u"wb")
        fp.write(b"hello" * 1000)
        assert not fp.close()

        self.runtest([u"testfiles/empty_dir", lf_dir.uc_name,
                      u"testfiles/empty_dir", lf_dir.uc_name])

    def test_empty_restore(self):
        u"""Make sure error raised when restore doesn't match anything"""
        self.backup(u"full", u"testfiles/dir1")
        self.assertRaises(CmdError, self.restore, u"this_file_does_not_exist")
        self.backup(u"inc", u"testfiles/empty_dir")
        self.assertRaises(CmdError, self.restore, u"this_file_does_not_exist")

    def test_remove_older_than(self):
        u"""Test removing old backup chains"""
        first_chain = self.backup(u"full", u"testfiles/dir1", current_time=10000)
        first_chain |= self.backup(u"inc", u"testfiles/dir2", current_time=20000)
        second_chain = self.backup(u"full", u"testfiles/dir1", current_time=30000)
        second_chain |= self.backup(u"inc", u"testfiles/dir3", current_time=40000)

        self.assertEqual(self.get_backend_files(), first_chain | second_chain)

        self.run_duplicity(options=[u"remove-older-than", u"35000", u"--force", self.backend_url])
        self.assertEqual(self.get_backend_files(), second_chain)

        # Now check to make sure we can't delete only chain
        self.run_duplicity(options=[u"remove-older-than", u"50000", u"--force", self.backend_url])
        self.assertEqual(self.get_backend_files(), second_chain)

    def test_piped_password(self):
        u"""Make sure that prompting for a password works"""
        self.set_environ(u"PASSPHRASE", None)
        self.backup(u"full", u"testfiles/empty_dir",
                    passphrase_input=[self.sign_passphrase, self.sign_passphrase])
        self.restore(passphrase_input=[self.sign_passphrase])


class OldFilenamesFinalTest(FinalTest):

    def setUp(self):
        super(OldFilenamesFinalTest, self).setUp()
        self.class_args.extend([u"--old-filenames"])


class ShortFilenamesFinalTest(FinalTest):

    def setUp(self):
        super(ShortFilenamesFinalTest, self).setUp()
        self.class_args.extend([u"--short-filenames"])

if __name__ == u"__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2015 Steve Tynor <steve.tynor@gmail.com>
# Copyright 2016 Thomas Harning Jr <harningt@gmail.com>
#                  - mirror/stripe modes
#                  - write error modes
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

#

from future import standard_library
standard_library.install_aliases()
import os
import os.path
import string
import sys
import urllib.request  # pylint: disable=import-error
import urllib.parse  # pylint: disable=import-error
import urllib.error  # pylint: disable=import-error
import json

import duplicity.backend
from duplicity.errors import BackendException
from duplicity import globals
from duplicity import log
from duplicity import util


class MultiBackend(duplicity.backend.Backend):
    u"""Store files across multiple remote stores. URL is a path to a local file
    containing URLs/other config defining the remote store"""

    # the stores we are managing
    __stores = []
    __affinities = {}

    # Set of known query paramaters
    __knownQueryParameters = frozenset([
        u'mode',
        u'onfail',
        u'subpath',
    ])

    # the mode of operation to follow
    # can be one of 'stripe' or 'mirror' currently
    __mode = u'stripe'
    __mode_allowedSet = frozenset([
        u'mirror',
        u'stripe',
    ])

    # the write error handling logic
    # can be one of the following:
    # * continue - default, on failure continues to next source
    # * abort - stop all further operations
    __onfail_mode = u'continue'
    __onfail_mode_allowedSet = frozenset([
        u'abort',
        u'continue',
    ])

    # sub path to dynamically add sub directories to backends
    # will be appended to the url value
    __subpath = u''

    # when we write in stripe mode, we "stripe" via a simple round-robin across
    # remote stores.  It's hard to get too much more sophisticated
    # since we can't rely on the backend to give us any useful meta
    # data (e.g. sizes of files, capacity of the store (quotas)) to do
    # a better job of balancing load across stores.
    __write_cursor = 0

    @staticmethod
    def get_query_params(parsed_url):
        # Reparse so the query string is available
        reparsed_url = urllib.parse.urlparse(parsed_url.geturl())
        if len(reparsed_url.query) == 0:
            return dict()
        try:
            queryMultiDict = urllib.parse.parse_qs(reparsed_url.query, strict_parsing=True)
        except ValueError as e:
            log.Log(_(u"MultiBackend: Could not parse query string %s: %s ")
                    % (reparsed_url.query, e),
                    log.ERROR)
            raise BackendException(u'Could not parse query string')
        queryDict = dict()
        # Convert the multi-dict to a single dictionary
        # while checking to make sure that no unrecognized values are found
        for name, valueList in list(queryMultiDict.items()):
            if len(valueList) != 1:
                log.Log(_(u"MultiBackend: Invalid query string %s: more than one value for %s")
                        % (reparsed_url.query, name),
                        log.ERROR)
                raise BackendException(u'Invalid query string')
            if name not in MultiBackend.__knownQueryParameters:
                log.Log(_(u"MultiBackend: Invalid query string %s: unknown parameter %s")
                        % (reparsed_url.query, name),
                        log.ERROR)
                raise BackendException(u'Invalid query string')

            queryDict[name] = valueList[0]
        return queryDict

    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)

        # Init each of the wrapped stores
        #
        # config file is a json formatted collection of values, one for
        # each backend.  We will 'stripe' data across all the given stores:
        #
        #  'url'  - the URL used for the backend store
        #  'env' - an optional list of enviroment variable values to set
        #      during the intialization of the backend
        #
        # Example:
        #
        # [
        #  {
        #   "url": "abackend://myuser@domain.com/backup",
        #   "env": [
        #     {
        #      "name" : "MYENV",
        #      "value" : "xyz"
        #     },
        #     {
        #      "name" : "FOO",
        #      "value" : "bar"
        #     }
        #    ]
        #  },
        #  {
        #   "url": "file:///path/to/dir"
        #  }
        # ]

        queryParams = MultiBackend.get_query_params(parsed_url)

        if u'mode' in queryParams:
            self.__mode = queryParams[u'mode']

        if u'onfail' in queryParams:
            self.__onfail_mode = queryParams[u'onfail']

        if self.__mode not in MultiBackend.__mode_allowedSet:
            log.Log(_(u"MultiBackend: illegal value for %s: %s")
                    % (u'mode', self.__mode), log.ERROR)
            raise BackendException(u"MultiBackend: invalid mode value")

        if self.__onfail_mode not in MultiBackend.__onfail_mode_allowedSet:
            log.Log(_(u"MultiBackend: illegal value for %s: %s")
                    % (u'onfail', self.__onfail_mode), log.ERROR)
            raise BackendException(u"MultiBackend: invalid onfail value")

        if u'subpath' in queryParams:
            self.__subpath = queryParams[u'subpath']

        try:
            with open(parsed_url.path) as f:
                configs = json.load(f)
        except IOError as e:
            log.Log(_(u"MultiBackend: Url %s")
                    % (parsed_url.geturl()),
                    log.ERROR)

            log.Log(_(u"MultiBackend: Could not load config file %s: %s ")
                    % (parsed_url.path, e),
                    log.ERROR)
            raise BackendException(u'Could not load config file')

        for config in configs:
            url = config[u'url'] + self.__subpath
            if sys.version_info.major == 2:
                url = url.encode(u'utf-8')
            log.Log(_(u"MultiBackend: use store %s")
                    % (url),
                    log.INFO)
            if u'env' in config:
                for env in config[u'env']:
                    log.Log(_(u"MultiBackend: set env %s = %s")
                            % (env[u'name'], env[u'value']),
                            log.INFO)
                    os.environ[env[u'name']] = env[u'value']

            store = duplicity.backend.get_backend(url)
            self.__stores.append(store)

            # Prefix affinity
            if u'prefixes' in config:
                if self.__mode == u'stripe':
                    raise BackendException(u"Multibackend: stripe mode not supported with prefix affinity.")
                for prefix in config[u'prefixes']:
                    log.Log(_(u"Multibackend: register affinity for prefix %s")
                            % prefix, log.INFO)
                if prefix in self.__affinities:
                    self.__affinities[prefix].append(store)
                else:
                    self.__affinities[prefix] = [store]

            # store_list = store.list()
            # log.Log(_("MultiBackend: at init, store %s has %s files")
            #         % (url, len(store_list)),
            #         log.INFO)

    def _eligible_stores(self, filename):
        if self.__affinities:
            matching_prefixes = [k for k in list(self.__affinities.keys()) if filename.startswith(k)]
            matching_stores = {store for prefix in matching_prefixes for store in self.__affinities[prefix]}
            if matching_stores:
                # Distinct stores with matching prefix
                return list(matching_stores)

        # No affinity rule or no matching store for that prefix
        return self.__stores

    def _put(self, source_path, remote_filename):
        # Store an indication of whether any of these passed
        passed = False

        # Eligibile stores for this action
        stores = self._eligible_stores(remote_filename)

        # Mirror mode always starts at zero
        if self.__mode == u'mirror':
            self.__write_cursor = 0

        first = self.__write_cursor
        while True:
            store = stores[self.__write_cursor]
            try:
                next = self.__write_cursor + 1
                if (next > len(stores) - 1):
                    next = 0
                log.Log(_(u"MultiBackend: _put: write to store #%s (%s)")
                        % (self.__write_cursor, store.backend.parsed_url.url_string),
                        log.DEBUG)
                store.put(source_path, remote_filename)
                passed = True
                self.__write_cursor = next
                # No matter what, if we loop around, break this loop
                if next == 0:
                    break
                # If in stripe mode, don't continue to the next
                if self.__mode == u'stripe':
                    break
            except Exception as e:
                log.Log(_(u"MultiBackend: failed to write to store #%s (%s), try #%s, Exception: %s")
                        % (self.__write_cursor, store.backend.parsed_url.url_string, next, e),
                        log.INFO)
                self.__write_cursor = next

                # If we consider write failure as abort, abort
                if self.__onfail_mode == u'abort':
                    log.Log(_(u"MultiBackend: failed to write %s. Aborting process.")
                            % (source_path),
                            log.ERROR)
                    raise BackendException(u"failed to write")

                # If we've looped around, and none of them passed, fail
                if (self.__write_cursor == first) and not passed:
                    log.Log(_(u"MultiBackend: failed to write %s. Tried all backing stores and none succeeded")
                            % (source_path),
                            log.ERROR)
                    raise BackendException(u"failed to write")

    def _get(self, remote_filename, local_path):
        # since the backend operations will be retried, we can't
        # simply try to get from the store, if not found, move to the
        # next store (since each failure will be retried n times
        # before finally giving up).  So we need to get the list first
        # before we try to fetch
        # ENHANCEME: maintain a cached list for each store
        stores = self._eligible_stores(remote_filename)

        for s in stores:
            list = s.list()
            if remote_filename in list:
                s.get(remote_filename, local_path)
                return
            log.Log(_(u"MultiBackend: failed to get %s to %s from %s")
                    % (remote_filename, local_path, s.backend.parsed_url.url_string),
                    log.INFO)
        log.Log(_(u"MultiBackend: failed to get %s. Tried all backing stores and none succeeded")
                % (remote_filename),
                log.ERROR)
        raise BackendException(u"failed to get")

    def _list(self):
        lists = []
        for s in self.__stores:
            globals.are_errors_fatal[u'list'] = (False, [])
            l = s.list()
            log.Notice(_(u"MultiBackend: %s: %d files")
                       % (s.backend.parsed_url.url_string, len(l)))
            if len(l) == 0 and duplicity.backend._last_exception:
                log.Warn(_(u"Exception during list of %s: %s"
                           % (s.backend.parsed_url.url_string,
                              util.uexc(duplicity.backend._last_exception))))
                duplicity.backend._last_exception = None
            lists.append(l)
        # combine the lists into a single flat list w/o duplicates via set:
        result = list({item for sublist in lists for item in sublist})
        log.Log(_(u"MultiBackend: combined list: %s")
                % (result),
                log.DEBUG)
        return result

    def _delete(self, filename):
        # Store an indication on whether any passed
        passed = False

        stores = self._eligible_stores(filename)

        # since the backend operations will be retried, we can't
        # simply try to get from the store, if not found, move to the
        # next store (since each failure will be retried n times
        # before finally giving up).  So we need to get the list first
        # before we try to delete
        # ENHANCEME: maintain a cached list for each store
        for s in stores:
            list = s.list()
            if filename in list:
                if hasattr(s, u'_delete_list'):
                    s._do_delete_list([filename, ])
                elif hasattr(s, u'_delete'):
                    s._do_delete(filename)
                passed = True
                # In stripe mode, only one item will have the file
                if self.__mode == u'stripe':
                    return
            log.Log(_(u"MultiBackend: failed to delete %s from %s")
                    % (filename, s.backend.parsed_url.url_string),
                    log.INFO)
        if not passed:
            log.Log(_(u"MultiBackend: failed to delete %s. Tried all backing stores and none succeeded")
                    % (filename),
                    log.ERROR)
#           raise BackendException("failed to delete")


duplicity.backend.register_backend(u'multi', MultiBackend)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
# Copyright 2013 Edgar Soldin
#                 - ssl cert verification, some robustness enhancements
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import range
import base64
import http.client
import os
import re
import shutil
import urllib.request  # pylint: disable=import-error
import urllib.parse  # pylint: disable=import-error
import urllib.error  # pylint: disable=import-error
import xml.dom.minidom

import duplicity.backend
from duplicity import globals
from duplicity import log
from duplicity import util
from duplicity.errors import BackendException, FatalBackendException


class CustomMethodRequest(urllib.request.Request):
    u"""
    This request subclass allows explicit specification of
    the HTTP request method. Basic urllib2.Request class
    chooses GET or POST depending on self.has_data()
    """
    def __init__(self, method, *args, **kwargs):
        self.method = method
        urllib.request.Request.__init__(self, *args, **kwargs)

    def get_method(self):
        return self.method


class VerifiedHTTPSConnection(http.client.HTTPSConnection):
    def __init__(self, *args, **kwargs):
        try:
            global socket, ssl
            import socket
            import ssl
        except ImportError:
            raise FatalBackendException(_(u"Missing socket or ssl python modules."))

        http.client.HTTPSConnection.__init__(self, *args, **kwargs)

        self.cacert_file = globals.ssl_cacert_file
        self.cacert_candidates = [u"~/.duplicity/cacert.pem",
                                  u"~/duplicity_cacert.pem",
                                  u"/etc/duplicity/cacert.pem"]
        # if no cacert file was given search default locations
        if not self.cacert_file:
            for path in self.cacert_candidates:
                path = os.path.expanduser(path)
                if (os.path.isfile(path)):
                    self.cacert_file = path
                    break

        # check if file is accessible (libssl errors are not very detailed)
        if self.cacert_file and not os.access(self.cacert_file, os.R_OK):
            raise FatalBackendException(_(u"Cacert database file '%s' is not readable.") %
                                        self.cacert_file)

    def connect(self):
        # create new socket
        sock = socket.create_connection((self.host, self.port),
                                        self.timeout)
        if self._tunnel_host:
            self.sock = sock
            self.tunnel()

        # python 2.7.9+ supports default system certs now
        if u"create_default_context" in dir(ssl):
            context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH,
                                                 cafile=self.cacert_file,
                                                 capath=globals.ssl_cacert_path)
            self.sock = context.wrap_socket(sock, server_hostname=self.host)
        # the legacy way needing a cert file
        else:
            if globals.ssl_cacert_path:
                raise FatalBackendException(
                    _(u"Option '--ssl-cacert-path' is not supported "
                      u"with python 2.7.8 and below."))

            if not self.cacert_file:
                raise FatalBackendException(_(u"""\
For certificate verification with python 2.7.8 or earlier a cacert database
file is needed in one of these locations: %s
Hints:
Consult the man page, chapter 'SSL Certificate Verification'.
Consider using the options --ssl-cacert-file, --ssl-no-check-certificate .""") %
                                            u", ".join(self.cacert_candidates))

            # wrap the socket in ssl using verification
            self.sock = ssl.wrap_socket(sock,
                                        cert_reqs=ssl.CERT_REQUIRED,
                                        ca_certs=self.cacert_file,
                                        )

    def request(self, *args, **kwargs):  # pylint: disable=method-hidden
        try:
            return http.client.HTTPSConnection.request(self, *args, **kwargs)
        except ssl.SSLError as e:
            # encapsulate ssl errors
            raise BackendException(u"SSL failed: %s" % util.uexc(e),
                                   log.ErrorCode.backend_error)


class WebDAVBackend(duplicity.backend.Backend):
    u"""Backend for accessing a WebDAV repository.

    webdav backend contributed in 2006 by Jesper Zedlitz <jesper@zedlitz.de>
    """

    u"""
    Request just the names.
    """
    listbody = u'<?xml version="1.0"?><D:propfind xmlns:D="DAV:"><D:prop><D:resourcetype/></D:prop></D:propfind>'

    u"""Connect to remote store using WebDAV Protocol"""
    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)
        self.headers = {u'Connection': u'keep-alive'}
        self.parsed_url = parsed_url
        self.digest_challenge = None
        self.digest_auth_handler = None

        self.username = parsed_url.username
        self.password = self.get_password()
        self.directory = self.sanitize_path(parsed_url.path)

        log.Info(_(u"Using WebDAV protocol %s") % (globals.webdav_proto,))
        log.Info(_(u"Using WebDAV host %s port %s") % (parsed_url.hostname,
                                                       parsed_url.port))
        log.Info(_(u"Using WebDAV directory %s") % (self.directory,))

        self.conn = None

    def sanitize_path(self, path):
        if path:
            foldpath = re.compile(u'/+')
            return foldpath.sub(u'/', path + u'/')
        else:
            return u'/'

    def getText(self, nodelist):
        rc = u""
        for node in nodelist:
            if node.nodeType == node.TEXT_NODE:
                rc = rc + node.data
        return rc

    def _retry_cleanup(self):
        self.connect(forced=True)

    def connect(self, forced=False):
        u"""
        Connect or re-connect to the server, updates self.conn
        # reconnect on errors as a precaution, there are errors e.g.
        # "[Errno 32] Broken pipe" or SSl errors that render the connection unusable
        """
        if not forced and self.conn \
                and self.conn.host == self.parsed_url.hostname:
            return

        log.Info(_(u"WebDAV create connection on '%s'") % (self.parsed_url.hostname))
        self._close()
        # http schemes needed for redirect urls from servers
        if self.parsed_url.scheme in [u'webdav', u'http']:
            self.conn = http.client.HTTPConnection(self.parsed_url.hostname, self.parsed_url.port)
        elif self.parsed_url.scheme in [u'webdavs', u'https']:
            if globals.ssl_no_check_certificate:
                self.conn = http.client.HTTPSConnection(self.parsed_url.hostname, self.parsed_url.port)
            else:
                self.conn = VerifiedHTTPSConnection(self.parsed_url.hostname, self.parsed_url.port)
        else:
            raise FatalBackendException(_(u"WebDAV Unknown URI scheme: %s") % (self.parsed_url.scheme))

    def _close(self):
        if self.conn:
            self.conn.close()

    def request(self, method, path, data=None, redirected=0):
        u"""
        Wraps the connection.request method to retry once if authentication is
        required
        """
        self._close()  # or we get previous request's data or exception
        self.connect()

        quoted_path = urllib.parse.quote(path, u"/:~")

        if self.digest_challenge is not None:
            self.headers[u'Authorization'] = self.get_digest_authorization(path)

        log.Info(_(u"WebDAV %s %s request with headers: %s ") % (method, quoted_path, self.headers))
        log.Info(_(u"WebDAV data length: %s ") % len(str(data)))
        self.conn.request(method, quoted_path, data, self.headers)
        response = self.conn.getresponse()
        log.Info(_(u"WebDAV response status %s with reason '%s'.") % (response.status, response.reason))
        # resolve redirects and reset url on listing requests (they usually come before everything else)
        if response.status in [301, 302] and method == u'PROPFIND':
            redirect_url = response.getheader(u'location', None)
            response.close()
            if redirect_url:
                log.Notice(_(u"WebDAV redirect to: %s ") % urllib.parse.unquote(redirect_url))
                if redirected > 10:
                    raise FatalBackendException(_(u"WebDAV redirected 10 times. Giving up."))
                self.parsed_url = duplicity.backend.ParsedUrl(redirect_url)
                self.directory = self.sanitize_path(self.parsed_url.path)
                return self.request(method, self.directory, data, redirected + 1)
            else:
                raise FatalBackendException(_(u"WebDAV missing location header in redirect response."))
        elif response.status == 401:
            response.read()
            response.close()
            self.headers[u'Authorization'] = self.get_authorization(response, quoted_path)
            log.Info(_(u"WebDAV retry request with authentification headers."))
            log.Info(_(u"WebDAV %s %s request2 with headers: %s ") % (method, quoted_path, self.headers))
            log.Info(_(u"WebDAV data length: %s ") % len(str(data)))
            self.conn.request(method, quoted_path, data, self.headers)
            response = self.conn.getresponse()
            log.Info(_(u"WebDAV response2 status %s with reason '%s'.") % (response.status, response.reason))

        return response

    def get_authorization(self, response, path):
        u"""
        Fetches the auth header based on the requested method (basic or digest)
        """
        try:
            auth_hdr = response.getheader(u'www-authenticate', u'')
            token, challenge = auth_hdr.split(u' ', 1)
        except ValueError:
            return None
        if token.split(u',')[0].lower() == u'negotiate':
            try:
                return self.get_kerberos_authorization()
            except ImportError:
                log.Warn(_(u"python-kerberos needed to use kerberos \
                          authorization, falling back to basic auth."))
                return self.get_basic_authorization()
            except Exception as e:
                log.Warn(_(u"Kerberos authorization failed: %s.\
                          Falling back to basic auth.") % e)
                return self.get_basic_authorization()
        elif token.lower() == u'basic':
            return self.get_basic_authorization()
        else:
            self.digest_challenge = self.parse_digest_challenge(challenge)
            return self.get_digest_authorization(path)

    def parse_digest_challenge(self, challenge_string):
        return urllib2.parse_keqv_list(urllib2.parse_http_list(challenge_string))

    def get_kerberos_authorization(self):
        import kerberos  # pylint: disable=import-error
        _, ctx = kerberos.authGSSClientInit(u"HTTP@%s" % self.conn.host)
        kerberos.authGSSClientStep(ctx, u"")
        tgt = kerberos.authGSSClientResponse(ctx)
        return u'Negotiate %s' % tgt

    def get_basic_authorization(self):
        u"""
        Returns the basic auth header
        """
        auth_string = u'%s:%s' % (self.username, self.password)
        return u'Basic %s' % base64.b64encode(auth_string.encode()).strip().decode()

    def get_digest_authorization(self, path):
        u"""
        Returns the digest auth header
        """
        u = self.parsed_url
        if self.digest_auth_handler is None:
            pw_manager = urllib.request.HTTPPasswordMgrWithDefaultRealm()
            pw_manager.add_password(None, self.conn.host, self.username, self.password)
            self.digest_auth_handler = urllib.request.HTTPDigestAuthHandler(pw_manager)

        # building a dummy request that gets never sent,
        # needed for call to auth_handler.get_authorization
        scheme = u.scheme == u'webdavs' and u'https' or u'http'
        hostname = u.port and u"%s:%s" % (u.hostname, u.port) or u.hostname
        dummy_url = u"%s://%s%s" % (scheme, hostname, path)
        dummy_req = CustomMethodRequest(self.conn._method, dummy_url)
        auth_string = self.digest_auth_handler.get_authorization(dummy_req,
                                                                 self.digest_challenge)
        return u'Digest %s' % auth_string

    def _list(self):
        response = None
        try:
            self.headers[u'Depth'] = u"1"
            response = self.request(u"PROPFIND", self.directory, self.listbody)
            del self.headers[u'Depth']
            # if the target collection does not exist, create it.
            if response.status == 404:
                response.close()  # otherwise next request fails with ResponseNotReady
                self.makedir()
                # just created an empty folder, so return empty
                return []
            elif response.status in [200, 207]:
                document = response.read()
                response.close()
            else:
                status = response.status
                reason = response.reason
                response.close()
                raise BackendException(u"Bad status code %s reason %s." % (status, reason))

            log.Debug(u"%s" % (document,))
            dom = xml.dom.minidom.parseString(document)
            result = []
            for href in dom.getElementsByTagName(u'd:href') + dom.getElementsByTagName(u'D:href'):
                filename = self.taste_href(href)
                if filename:
                    result.append(filename)
            return result
        except Exception as e:
            raise e
        finally:
            if response:
                response.close()

    def makedir(self):
        u"""Make (nested) directories on the server."""
        dirs = self.directory.split(u"/")
        # url causes directory to start with /, but it might be given
        # with or without trailing / (which is required)
        if dirs[-1] == u'':
            dirs = dirs[0:-1]
        for i in range(1, len(dirs)):
            d = u"/".join(dirs[0:i + 1]) + u"/"

            self.headers[u'Depth'] = u"1"
            response = self.request(u"PROPFIND", d)
            del self.headers[u'Depth']

            log.Info(u"Checking existence dir %s: %d" % (d, response.status))

            if response.status == 404:
                log.Info(_(u"Creating missing directory %s") % d)

                res = self.request(u"MKCOL", d)
                if res.status != 201:
                    raise BackendException(_(u"WebDAV MKCOL %s failed: %s %s") %
                                           (d, res.status, res.reason))

    def taste_href(self, href):
        u"""
        Internal helper to taste the given href node and, if
        it is a duplicity file, collect it as a result file.

        @return: A matching filename, or None if the href did not match.
        """
        raw_filename = self.getText(href.childNodes).strip()
        parsed_url = urllib.parse.urlparse(urllib.parse.unquote(raw_filename))
        filename = parsed_url.path
        log.Debug(_(u"WebDAV path decoding and translation: "
                  u"%s -> %s") % (raw_filename, filename))

        # at least one WebDAV server returns files in the form
        # of full URL:s. this may or may not be
        # according to the standard, but regardless we
        # feel we want to bail out if the hostname
        # does not match until someone has looked into
        # what the WebDAV protocol mandages.
        if parsed_url.hostname is not None \
           and not (parsed_url.hostname == self.parsed_url.hostname):
            m = u"Received filename was in the form of a "\
                u"full url, but the hostname (%s) did "\
                u"not match that of the webdav backend "\
                u"url (%s) - aborting as a conservative "\
                u"safety measure. If this happens to you, "\
                u"please report the problem"\
                u"" % (parsed_url.hostname,
                       self.parsed_url.hostname)
            raise BackendException(m)

        if filename.startswith(self.directory):
            filename = filename.replace(self.directory, u'', 1)
            return filename
        else:
            return None

    def _get(self, remote_filename, local_path):
        url = self.directory + util.fsdecode(remote_filename)
        response = None
        try:
            target_file = local_path.open(u"wb")
            response = self.request(u"GET", url)
            if response.status == 200:
                # data=response.read()
                shutil.copyfileobj(response, target_file)
                # import hashlib
                # log.Info("WebDAV GOT %s bytes with md5=%s" %
                # (len(data),hashlib.md5(data).hexdigest()) )
                assert not target_file.close()
                response.close()
            else:
                status = response.status
                reason = response.reason
                response.close()
                raise BackendException(_(u"WebDAV GET Bad status code %s reason %s.") %
                                       (status, reason))
        except Exception as e:
            raise e
        finally:
            if response:
                response.close()

    def _put(self, source_path, remote_filename):
        url = self.directory + util.fsdecode(remote_filename)
        response = None
        try:
            source_file = source_path.open(u"rb")
            response = self.request(u"PUT", url, source_file)
            # 200 is returned if a file is overwritten during restarting
            if response.status in [200, 201, 204]:
                response.read()
                response.close()
            else:
                status = response.status
                reason = response.reason
                response.close()
                raise BackendException(_(u"WebDAV PUT Bad status code %s reason %s.") %
                                       (status, reason))
        except Exception as e:
            raise e
        finally:
            if response:
                response.close()

    def _delete(self, filename):
        url = self.directory + util.fsdecode(filename)
        response = None
        try:
            response = self.request(u"DELETE", url)
            if response.status in [200, 204]:
                response.read()
                response.close()
            else:
                status = response.status
                reason = response.reason
                response.close()
                raise BackendException(_(u"WebDAV DEL Bad status code %s reason %s.") %
                                       (status, reason))
        except Exception as e:
            raise e
        finally:
            if response:
                response.close()


duplicity.backend.register_backend(u"http", WebDAVBackend)
duplicity.backend.register_backend(u"https", WebDAVBackend)
duplicity.backend.register_backend(u"webdav", WebDAVBackend)
duplicity.backend.register_backend(u"webdavs", WebDAVBackend)
duplicity.backend.uses_netloc.extend([u'http', u'https', u'webdav', u'webdavs'])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

u"""
Miscellaneous utilities.
"""

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import isinstance
from builtins import map
from builtins import object
from builtins import str

import errno
import os
import string
import sys
import traceback
import atexit

from duplicity import tarfile
import duplicity.globals as globals
import duplicity.log as log

try:
    # For paths, just use path.name/uname rather than converting with these
    from os import fsencode, fsdecode
except ImportError:
    # Most likely Python version < 3.2, so define our own fsencode/fsdecode.
    # These are functions that encode/decode unicode paths to filesystem encoding,
    # but the cleverness is that they handle non-unicode characters on Linux
    # There is a *partial* backport to python available here:
    # https://github.com/pjdelport/backports.os/blob/master/src/backports/os.py
    # but if it cannot be trusted for full-circle translation, then we may as well
    # just read and store the bytes version of the path as path.name before
    # creating the unicode version (for path matching etc) and ensure that in
    # real-world usage (as opposed to testing) we create the path objects from a
    # bytes string.
    # ToDo: Revisit this once we drop Python 2 support/the backport is complete

    def fsencode(unicode_filename):
        u"""Convert a unicode filename to a filename encoded in the system encoding"""
        # For paths, just use path.name rather than converting with this
        # If we are not doing any cleverness with non-unicode filename bytes,
        # encoding to system encoding is good enough
        return unicode_filename.encode(sys.getfilesystemencoding(), u"replace")

    def fsdecode(bytes_filename):
        u"""Convert a filename encoded in the system encoding to unicode"""
        # For paths, just use path.uc_name rather than converting with this
        # If we are not doing any cleverness with non-unicode filename bytes,
        # decoding using system encoding is good enough. Use "ignore" as
        # Linux paths can contain non-Unicode characters
        return bytes_filename.decode(globals.fsencoding, u"replace")


def exception_traceback(limit=50):
    u"""
    @return A string representation in typical Python format of the
            currently active/raised exception.
    """
    type, value, tb = sys.exc_info()

    lines = traceback.format_tb(tb, limit)
    lines.extend(traceback.format_exception_only(type, value))

    msg = u"Traceback (innermost last):\n"
    if sys.version_info.major >= 3:
        msg = msg + u"%-20s %s" % (str.join(u"", lines[:-1]), lines[-1])
    else:
        msg = msg + u"%-20s %s" % (string.join(lines[:-1], u""), lines[-1])

    if sys.version_info.major < 3:
        return msg.decode(u'unicode-escape', u'replace')
    return msg


def escape(string):
    u"Convert a (bytes) filename to a format suitable for logging (quoted utf8)"
    string = fsdecode(string).encode(u'unicode-escape', u'replace')
    return u"'%s'" % string.decode(u'utf8', u'replace')


def uindex(index):
    u"Convert an index (a tuple of path parts) to unicode for printing"
    if index:
        return os.path.join(*list(map(fsdecode, index)))
    else:
        return u'.'


def uexc(e):
    u"""Returns the exception message in Unicode"""
    # Exceptions in duplicity often have path names in them, which if they are
    # non-ascii will cause a UnicodeDecodeError when implicitly decoding to
    # unicode.  So we decode manually, using the filesystem encoding.
    # 99.99% of the time, this will be a fine encoding to use.
    if e.args:
        # Find arg that is a string
        for m in e.args:
            if isinstance(m, str):
                # Already unicode
                return m
            elif isinstance(m, bytes):
                # Encoded, likely in filesystem encoding
                return fsdecode(m)
    else:
        return u''


def maybe_ignore_errors(fn):
    u"""
    Execute fn. If the global configuration setting ignore_errors is
    set to True, catch errors and log them but do continue (and return
    None).

    @param fn: A callable.
    @return Whatever fn returns when called, or None if it failed and ignore_errors is true.
    """
    try:
        return fn()
    except Exception as e:
        if globals.ignore_errors:
            log.Warn(_(u"IGNORED_ERROR: Warning: ignoring error as requested: %s: %s")
                     % (e.__class__.__name__, uexc(e)))
            return None
        else:
            raise


class BlackHoleList(list):

    def append(self, x):
        pass


class FakeTarFile(object):
    debug = 0

    def __iter__(self):
        return iter([])

    def close(self):
        pass


def make_tarfile(mode, fp):
    # We often use 'empty' tarfiles for signatures that haven't been filled out
    # yet.  So we want to ignore ReadError exceptions, which are used to signal
    # this.
    try:
        tf = tarfile.TarFile(u"arbitrary", mode, fp)
        # Now we cause TarFile to not cache TarInfo objects.  It would end up
        # consuming a lot of memory over the lifetime of our long-lasting
        # signature files otherwise.
        tf.members = BlackHoleList()
        return tf
    except tarfile.ReadError:
        return FakeTarFile()


def get_tarinfo_name(ti):
    # Python versions before 2.6 ensure that directories end with /, but 2.6
    # and later ensure they they *don't* have /.  ::shrug::  Internally, we
    # continue to use pre-2.6 method.
    if ti.isdir() and not ti.name.endswith(r"/"):
        return ti.name + r"/"
    else:
        return ti.name


def ignore_missing(fn, filename):
    u"""
    Execute fn on filename.  Ignore ENOENT errors, otherwise raise exception.

    @param fn: callable
    @param filename: string
    """
    try:
        fn(filename)
    except OSError as ex:
        if ex.errno == errno.ENOENT:
            pass
        else:
            raise


@atexit.register
def release_lockfile():
    if globals.lockfile:
        log.Debug(_(u"Releasing lockfile %s") % globals.lockpath)
        try:
            globals.lockfile.release()
        except Exception:
            pass


def copyfileobj(infp, outfp, byte_count=-1):
    u"""Copy byte_count bytes from infp to outfp, or all if byte_count < 0

    Returns the number of bytes actually written (may be less than
    byte_count if find eof.  Does not close either fileobj.

    """
    blocksize = 64 * 1024
    bytes_written = 0
    if byte_count < 0:
        while 1:
            buf = infp.read(blocksize)
            if not buf:
                break
            bytes_written += len(buf)
            outfp.write(buf)
    else:
        while bytes_written + blocksize <= byte_count:
            buf = infp.read(blocksize)
            if not buf:
                break
            bytes_written += len(buf)
            outfp.write(buf)
        buf = infp.read(byte_count - bytes_written)
        bytes_written += len(buf)
        outfp.write(buf)
    return bytes_written


def which(program):
    u"""
    Return absolute path for program name.
    Returns None if program not found.
    """

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.path.isabs(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)  # @UnusedVariable
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.getenv(u"PATH").split(os.pathsep):
            path = path.strip(u'"')
            exe_file = os.path.abspath(os.path.join(path, program))
            if is_exe(exe_file):
                return exe_file

    return None
